
/* 
   (c) COPYRIGHT URI/MIT 1999
   Please read the full copyright statement in the file COPYRIGHT.

   Authors:
      jhrg,jimg		James Gallagher (jgallagher@gso.uri.edu)
*/

/** Simple interface for loaddods error reporting. This uses two global matlab
    variables, dods_err and dods_err_msg. dods_err is a scalar and is
    initailly 0 to indicate that no error has been found. When an error is
    found it is set to 1. Callers of loaddods must test this variable to
    determine if an error has occurred. If dods_err is 1 then an error
    message will be contained in dods_err_msg (a Matlab vector).

    This interface defines nine functions: err_init() is used to initialize
    the error `system'. The function err_msg() is used to set the dods_err
    and dods_err_msg variables. Messages are terminated by a \n character. If
    err_msg() is called more than once before the program exits, the messages
    will be appended to dods_err_msg until the size of that char array
    becomes greater than MSG_STR (4064 characters). At that point more calls
    to err_msg simply result in the `Too many messages.' string being append
    to dods_err_msg. 

    A companion function msg() exists which can be called to send messages to
    stderr (so that users can see them right away; useful for debuging
    instrumentation and informational messages). The initialization function
    err_init() can be called so that the err_msg() function behaves
    identically to the msg() function. This provides complete backward
    compatibility with the old error reporting scheme which sent all messages
    to stderr.

    If the compatibility mode is used, callers of loaddods must be savvy and
    not look for dods_err and/or dods_err_msg since they may not exist. 

    If dods_err is false (0), the contents of dods_err_msg are undefined.

    @author jhrg */

/* 
 * $Log: error.c,v $
 * Revision 1.3.2.1  1999/11/02 23:49:37  jimg
 * Added instrumentation
 *
 * Revision 1.3  1999/07/24 00:09:29  jimg
 * Added include of config_writeval
 *
 * Revision 1.2  1999/04/30 17:06:57  jimg
 * Merged with no-gnu and release-2-24
 *
 * Revision 1.1.2.1  1999/04/29 20:10:26  jimg
 * Added
 * 
 */

#include "config_writeval.h"

#include <stdarg.h>
#include <stdio.h>

#ifdef TEST_LOADDODS
#define matlab 0
#else
#define matlab 1
#endif

#if matlab
#include <mex.h>
#endif

#define MSG_STR 4064		/* MSG_STR + OVERFLOW = 4K */
#define OVERFLOW 32		/* We need 25, but why not make it 2^5? */

#define OVERFLOW_MSG "Too many error messages.\n"

static int _comp_mode = 0;
static int _dods_err = 0;
static char _dods_err_msg[MSG_STR+OVERFLOW] = "";
static int _overflow_dods_err_msg = 0;

/** Internal function. Glue the newest message onto other, older, messages.
    If the message limit is run over, tack a `Too many error messages'
    message on. Messages all are terminated by a newline.

    @param dem char * to the global error message cache.
    @param msg const char * to the newest message. */
static void
_append_new_msg(char *dem, const char *msg)
{
    if (_overflow_dods_err_msg)
	return;

    if (strlen(dem) + strlen(msg) > MSG_STR) {
	_overflow_dods_err_msg = 1;
	strcat(dem, OVERFLOW_MSG);
    }
    else {
	strcat(dem, msg);
    }
}	

/** Take the value of _err_state and dump it into the Matlab scalar dods_err.
 */ 
static void
_intern_dods_err(int _err_state)
{
    double dummy = 0;
    double dde = _err_state;
#if matlab
    int error = mexPutFull("dods_err", 1, 1, &dde, &dummy);

    if (error)
	mexErrMsgTxt("Internal Error in the reporting system (1)\n\
Please report this error to support@unidata.ucar.edu");
#endif
}

/** Take the value of _msg and dump it into dods_err_msg. */
static void
_intern_msg(char *_msg)
{
#if matlab
    int error = 0;
    Matrix *pm = mxCreateString(_msg);

    mxSetName(pm, "dods_err_msg");
    error = mexPutMatrix(pm);
    if (error)
	mexErrMsgTxt("Internal Error in the reporting system (1)\n\
Please report this error to support@unidata.ucar.edu");
#endif
}

/** Initialize the error reporting subsystem of loaddods. 
    
    @param compat_mode If 1, use the compatibility mode where all error
    messages are printed to stderr. If 0, error messages are sent to the
    Matlab `string' variable (ML 4 compatible) and dods_err is set to 1. */
int
err_init(int compat_mode)
{
    _comp_mode = compat_mode;	/* Old behavior or new behavior? */
    if (!_comp_mode) {
	/* create/initialize the dods_err, etc. Since Matlab keeps the
	   mexFunction stuff around across invocations, we must reinit the
	   static variables in addition to the Matlab variables. */
	_dods_err = 0;
	_intern_dods_err(_dods_err);
	_dods_err_msg[0] = '\0';
	_overflow_dods_err_msg = 0;
    }
}

/** Write a message to standard error. The matlab variables dods_err and
    dods_err_msg are not affected.

    @param m Write the messsage #m# to standard error. */
void 
msg(const char *m, ...) 
{
    va_list arg;
    va_start(arg, m);

    vfprintf(stderr, m, arg);
}

/** Write a message to the Matlab variable dods_err_msg. Also set the Matlab
    variable dods_err to 1. If err_msg() has been called since the last call
    to err_init(), this message is appended and the state of dods_err (which
    is already 1) is not changed.

    @param m Write the message #m# to the dods_err_msg Matlab variable. */
void 
err_msg(const char *m, ...) 
{
    va_list arg;
    va_start(arg, m);

    if (_comp_mode)
	vfprintf(stderr, m, arg);
#if 0
	msg(m, arg);
#endif
    else {
	char s[MSG_STR];
	vsprintf(s, m, arg);

	_dods_err = 1;
	_intern_dods_err(_dods_err);

	_append_new_msg(_dods_err_msg, s);
	_intern_msg(_dods_err_msg);
    }
}

#ifdef TEST

/* 
   Call with a.out <0|1> <message 0> ... <message N>. Each call to err_msg
   include three integer arguments.
*/

int
main(int argc, char *argv[])
{
    int i;

    err_init(atoi(argv[1]));

    i = 2;
    while (i < argc) {
	err_msg(argv[i], i, i, i);
	i++;
    };

    printf("dods_err: %d, dods_err_msg: `%s'\n", _dods_err, _dods_err_msg);
}

#endif
